;******************************************************************************
;
;    This program is free software; you can redistribute it and/or modify
;    it under the terms of the GNU General Public License as published by
;    the Free Software Foundation; either version 2 of the License, or
;    (at your option) any later version.
;
;    This program is distributed in the hope that it will be useful,
;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;    GNU General Public License for more details.
;
;    You should have received a copy of the GNU General Public License
;    along with this program; if not, write to the Free Software
;    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
;
;    I (Simon Wood) can be contacted via email - unilink@mungewell.org
;    or via post - 14 Kirklands Close, Baildon, West Yorkshire, England.
;
;******************************************************************************
; Functions for GNUnilink - these have to be placed at start of code as the 
; PIC12 series can only 'Call' to lower addresses

;******************************************************************************
; Send a Slave Break - tells the Master that you have something to tell them...

; In this function GNUnilink looks for a period when the data bus is low for 
; more than 6ms and then waits for it to be high for 2ms. Once this is seen it 
; knows were in the middle of an unused 'slave phase' and it then forces the 
; bus low for 4ms.

; NOTE - whilst GNUnilink is in this state it is unable to receive other 
; commands from the master

IssueSlaveBreak:
	bcf	SomeFlags, SBRQ			; clear flag

	IFDEF	DEBUG
	movlw	0xAF				; 0xAF: IssueSlaveBreak
	call	SendSerial
	ENDIF

IssueSlaveBreak_Restart:
	movlw	60				; 6ms (60 * 100us)
	movwf	Count

BreakLoop1:
	movlw	delay100us			; 100us 
	movwf	Count2

BreakLoop2:
	btfsc	OUT_PORT, DataBit		; Data low ('Master phase')
	goto	IssueSlaveBreak_Restart

	goto	BreakLoop3			; NOP
BreakLoop3:

	decfsz	Count2, 1			; Reduce Count2
	goto	BreakLoop2

	decfsz	Count, 1			; Reduce count 
	goto	BreakLoop1

	IFDEF	SERIAL
	bcf	INTCON, GIE
	ENDIF

BreakLoop4:
	IFDEF	NASTY_HACK
	movlw	40				; rely on timing. 
	ELSE
	movlw	20				; 2ms (20 * 100us)
	ENDIF
	movwf	Count

BreakLoop5:
	movlw	delay100us
	movwf	Count2

BreakLoop6:
	IFNDEF	NASTY_HACK
	btfss	OUT_PORT, DataBit		; Data high for 2ms 
	goto	BreakLoop4			; (moving into 'Slave Phase')
	ENDIF

	goto	BreakLoop7			; NOP
BreakLoop7:

	decfsz	Count2, 1			; Reduce Count2
	goto	BreakLoop6

	decfsz	Count, 1			; Reduce count 
	goto	BreakLoop5

	IFDEF PIC_12
	movlw	0xfe				; Enable data pin
	tris	0
	ENDIF

	IFDEF PIC_16
	bsf	STATUS, RP0			; Enable data pin as output
	bcf	OUT_ENA, DataBit
	bcf	STATUS, RP0
	ENDIF

	bcf	OUT_PORT, DataBit		; Force Data low.... 4ms

	movlw	40
	movwf	Count

BreakLoop9:
	movlw	delay100us
	movwf	Count2

BreakLoop10:
	goto	BreakLoop11			; 2 * NOP
BreakLoop11:
	goto	BreakLoop12			; 2 * NOP
BreakLoop12:

	decfsz	Count2, 1
	goto	BreakLoop10

	decfsz	Count, 1
	goto	BreakLoop9

	IFDEF PIC_12
	movlw	0xff				; Tristate data pin
	tris	0
	ENDIF

	IFDEF PIC_16
	bsf	STATUS, RP0			; Tristate data pin
	bsf	OUT_ENA, DataBit
	bcf	STATUS, RP0
	ENDIF

	IFDEF	SERIAL
	bsf	INTCON, GIE
	ENDIF

	return

;******************************************************************************
; Receive a byte - received byte is placed in W register.

; Input is sampled between 1 cycle before and 3 cycles after Clock rising edge,
; depending on the asynchronous alignment with the Master

ReadByte:
	clrf	DataVal

	movlw	0x08
	movwf	Count

	bcf	STATUS, C			; Ensure carry bit is 0

ReadLoop1:
	rlf	DataVal, 1
	
ReadLoop2:
	btfsc	SomeFlags, SBRQ			; SlaveBreak requested?
	goto	Receive_DoSlaveBreak

	btfss	OUT_PORT, ClockBit		; Wait for clock going high
	goto	ReadLoop2

	IFDEF	SERIAL
	bcf	INTCON, GIE
	ENDIF

ReadLoop3:
;	movf	OUT_PORT, 0			; Sample input

	btfsc	OUT_PORT, ClockBit		; Wait for clock going low
	goto	ReadLoop3

;	movwf	Count2				; (inverted) input
;	btfss	Count2, DataBit
	btfss	OUT_PORT, DataBit
	bsf	DataVal, 0
	
	decfsz	Count, 1			; Do next bit, if any left...
	goto	ReadLoop1

	IFDEF	SERIAL
	bsf	INTCON, GIE
	ENDIF

	IFDEF	DEBUG
	movf	DataVal, 0
	call	SendSerial
	ENDIF

	movf	DataVal, 0
	return

;******************************************************************************
; Send a byte - byte to ouput is in W register.

; This is the piece of code which is the most time critcal.
; Worst case the outpit bit is set a maximum of 7 cycles after Clock goes High
; and held until the Clock goes high again

; this hopefully gives enough time for the master to register it.... a LOT 
; dodgy on slow systems.

; We may have to reverse the logic on this so that the output is set before the 
; Clock goes high and changed after the Clock goes low. This will improve the i
; set up, but reduce the hold time.

SendByte:
	movwf	DataVal				; Store byte to output

	addwf	Checksum, 1			; Compute Checksum

	IFDEF	DEBUG
	call	SendSerial
	ENDIF

	IFDEF PIC_12
	movlw	0xfe				; Enable data pin
	tris	0
	ENDIF

	IFDEF PIC_16
	bsf	STATUS, RP0			; Enable data pin
	bcf	OUT_ENA, DataBit
	bcf	STATUS, RP0
	ENDIF

	IFDEF	SERIAL
	bcf	INTCON, GIE
	ENDIF
	
	movlw	0x08
	movwf	Count

	bcf	OUT_PORT, DataBit		; remember it's INVERTED

SendLoop1:
	btfsc	DataVal, 7
	goto	SendBitHigh

SendLoop2:
	btfss	OUT_PORT, ClockBit		; Wait for clock going high
	goto	SendLoop2

	bsf	OUT_PORT, DataBit

	goto	SendBitSent

SendBitHigh:
	btfss	OUT_PORT, ClockBit		; Wait for clock going high
	goto	SendBitHigh

	bcf	OUT_PORT, DataBit

SendBitSent:
	rlf	DataVal, 1

SendLoop3:
	btfsc	OUT_PORT, ClockBit		; Wait for clock going low
	goto	SendLoop3

	decfsz	Count, 1			; Do next bit, if any left...
	goto	SendLoop1

	movlw	3				; just a little data hold
	movwf	Count				; 2+(n-1)*3+2 clocks
SendLoop4:					; n=3: 10 clocks = 4us
	decfsz	Count, F
	goto	SendLoop4



;SendLoop1:
;	bcf	OUT_PORT, DataBit		; remember it's INVERTED

;SendLoop2:
;	btfss	OUT_PORT, ClockBit		; Wait for clock going high
;	goto	SendLoop2

;	btfss	DataVal, 7
;	bsf	OUT_PORT, DataBit

;	rlf	DataVal, 1

;SendLoop3:
;	btfsc	OUT_PORT, ClockBit		; Wait for clock going low
;	goto	SendLoop3

;	decfsz	Count, 1			; Do next bit, if any left...
;	goto	SendLoop1

;	goto	SendLoop4			; just a little data hold
;SendLoop4:

	IFDEF PIC_12
	movlw	0xff				; Tristate data pin
	tris	0
	ENDIF

	IFDEF PIC_16
	bsf	STATUS, RP0			; Trisate data pin
	bsf	OUT_ENA, DataBit
	bcf	STATUS, RP0
	ENDIF
	
	IFDEF	SERIAL
	bsf	INTCON, GIE
	ENDIF

	return

;******************************************************************************
; SendSerial - Send value in W out on PORTB.0 as RS232 TxD line

	IFDEF	DEBUG
SendSerial:
	movwf	SerByte

	movlw	9			; 8 bits + start bit
	movwf	Count
	bcf	STATUS, C		; start bit polarity
SendSer1:
	rrf	SerByte, 1
	btfsc	SerByte, 7
	goto	SendSer2
	nop
	bcf	PORTB, 0
	goto	SendSer3
SendSer2:
	bsf	PORTB, 0
	goto	SendSer3		; to maintain exact timing
SendSer3:
	movlw	4
	movwf	Count3			; 9 instruction cycles
SendSer4:
	decfsz	Count3, 1
	goto	SendSer4		; 2 + (CountValue-1)*3 instruction cycles

	decfsz	Count, 1
	goto	SendSer1		; 3 cycles; 2 for last bit
	nop

	nop
	nop
	nop
	nop

	bsf	PORTB, 0		; idle

	return
	ENDIF

;******************************************************************************
; Bit frig - works out which bit to set in the response to Master Poll

; W register is input of which stage you are on (0x00, 0x20, 0x30, 0x40 etc)
; and is returned with the byte to write (0x00 if wrong stage).

Bit_Frig:
	xorwf	MyBit, 0
	andlw	0xe0				; Strip off low bits

	btfsc	STATUS, Z			; Do we have a hit?
	goto	Bit_Frig_Hit

	movlw	0x00
	return

Bit_Frig_Hit:
	btfss	MyBit, 4			; Do we need to swap nybbles?
	goto	Bit_Frig_Swap

	movf	MyBit, 0
	andlw	0x0F
	return

Bit_Frig_Swap:
	swapf	MyBit, 0
	andlw	0xF0
	return

;******************************************************************************
; DB_WritePrep - prepare a byte write to the Display buffer

; The FSR register will be loaded correctly so that the desired value can be
; written into the INDF register after DB_WritePrep returns
; DisplayW pointer is set to point to the next free location

DB_WritePrep:
	movf	DisplayW, 0			; Load indirect pointer
	movwf	FSR

	incf	DisplayW, 1			; Set DisplayW to point to next byte
	movf	DisplayW, 0
	xorlw	DisplayEnd+1
	btfss	STATUS, Z
	goto	DB_WritePrep1
	movlw	Display0
	movwf	DisplayW

DB_WritePrep1:
	return

;******************************************************************************
; DB_ReadPrep - prepare a byte read from the Display buffer

; The FSR register will be loaded correctly so that the desired value can be
; read from the INDF register after DB_ReadPrep returns
; DisplayR pointer is set to point to the next location

DB_ReadPrep:
	movf	DisplayR, 0			; Load indirect pointer
	movwf	FSR

	incf	DisplayR, 1			; Set DisplayR to point to next byte
	movf	DisplayR, 0
	xorlw	DisplayEnd+1
	btfss	STATUS, Z
	goto	DB_ReadPrep1
	movlw	Display0
	movwf	DisplayR

DB_ReadPrep1:
	return

;******************************************************************************
; EEWrite - Write value in W to EEADR; don't wait until write finished

EEWrite:
	bsf	STATUS, RP0			; Bank 1
	movwf	EEDATA

	bcf	STATUS, RP0			; Bank 0
	bcf	PIR1, EEIF			; flag to determine write end
	bsf	STATUS, RP0			; EECONx are located in Bank 1
	bcf	INTCON, GIE			; avoid sequence to be interrupted
	bsf	EECON1, WREN
	movlw	0x55				; EEPROM write sequence
	movwf	EECON2
	movlw	0xAA
	movwf	EECON2
	bsf	EECON1, WR			; start write operation
	bcf	EECON1, WREN
;	bsf	INTCON, GIE			; re-enable interrupts - nein, weil EEWrite auch im Int aufgerufen wird!
	bcf	STATUS, RP0			; switch back to Bank 0

	return


EEWriteLoop1:
	btfss	EECON1, EEIF			; bit is set when write finished
	goto	EEWriteLoop1

	bcf	EECON1, WREN
	bcf	STATUS, RP0			; switch back to Bank 0
	incf	EEADR, 1

